#include <Wire.h>
#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x3F, 16, 2);

#include <NeoSWSerial.h>
NeoSWSerial ss(10, 11);

#include <TinyGPS++.h>
TinyGPSPlus gps;

#include <SerialRAM.h>
SerialRAM ram;

#define maxmode 23

struct nonvol_t { //structure that stores settings and status that should be nonvolatile
  byte backlt;    //backlight
  byte mode;
  float lat1;
  float lng1;
  uint64_t trip1;
  uint64_t trip2;
  uint64_t trip3;
  uint64_t trip4;
  uint64_t odometer;
  byte sevseg;
  uint32_t hourmeter; //accually stores total on time in seconds
  float last_lat1;
  float last_lng1;
  byte checksum;
};
union nonvolU {
  nonvol_t s;
  uint8_t b[sizeof(nonvol_t)];
};
nonvolU stat;

byte currentKey = 0;
byte ramcycle = 0;
byte lastMode = 255;
long pmillis; //for integration
unsigned long lastMillis = 0;
bool dispen = 1; //display (7 segment) enable
float bvolt;
uint32_t opening_hourmeter;
float opening_last_lat1;
float opening_last_lng1;
uint64_t tripPwrOn=0;

float highscore=0; //the highest speed seen since being powered on


void setup() {
  Serial.begin(9600);
  for (byte i = 0; i <= 7; i++) {
    pinMode(i + 2, OUTPUT);  //7 segment pins
  }
  for (byte i = 14; i <= 17; i++) {
    pinMode(i, INPUT_PULLUP);  //4 user buttons
  }
  pinMode(13, OUTPUT); digitalWrite(13, HIGH);
  pinMode(12, OUTPUT); digitalWrite(12, HIGH);
  analogReference(INTERNAL);

  ram.begin();
  //Serial.println(F("EERAM DUMP"));for(int j=0;j<2048;j++){if(j%10==0){Serial.println();Serial.print(j);Serial.print('\t');}Serial.print(ram.read(j));Serial.print('\t');}
  lcd.init(); lcd.noBacklight();
  ss.begin(9600);

  //stat.s.odometer=20*1609340000LL;
  stat.s.lat1 = 42.148109;
  stat.s.lng1 = -83.752826;

  stat.s.sevseg = 255;
  for (byte n = 0; n <= 9; n++) {
    disp7(11 * n); delay(70);
  }
  dispen = 0;

  if (userButtons() == 10) {
    lcd.backlight();
    lcd.print(F("Hold 10 to reset"));
    delay(10000);
    if (userButtons() == 10) {
      saveSettings();
      saveSettings();
      saveSettings();
      saveSettings();
    }
  }
  readSettings();
  opening_hourmeter = stat.s.hourmeter;
  opening_last_lat1 = stat.s.last_lat1;
  opening_last_lng1 = stat.s.last_lng1;
  applySettings();
  digitalWrite(12, LOW); digitalWrite(13, LOW);
  lcd.print(F("R ")); lcd.print(ramcycle);
  delay(700);
}

void loop() {
  while (ss.available()) {
    gps.encode(ss.read());
    yield();
  } integrateSpeed();
  //disp7((byte)gps.speed.mph());
  //disp7(stat.s.mode);
  bvolt = (float)0.005263671 * analogRead(A6) * (1 - .99) + bvolt * .99;
  currentKey = userButtons();

  if (gps.location.isValid() != 1 || gps.location.age() > 5000) {
    dispen = 0; digitalWrite(13, HIGH); //enable the display and turn off decimal point
  } else {
    dispen = 1; digitalWrite(13, LOW);
    stat.s.last_lat1 = gps.location.lat(); stat.s.last_lng1 = gps.location.lng();
    if(highscore<=gps.speed.mph()){highscore=gps.speed.mph();}
  }

  switch (currentKey) {
    case 2: stat.s.mode++; if (stat.s.mode > maxmode) {
        stat.s.mode = 0;
      } break;
    case 4: stat.s.mode--; if (stat.s.mode > maxmode) {
        stat.s.mode = maxmode;
      } break;
  }

  if (stat.s.mode != lastMode) {
    lcd.clear(); lastMode = stat.s.mode; lastMillis = 0;
    if (stat.s.mode != 0) {
      lcd.print(stat.s.mode);
      lcd.print(' ');
    }
    switch (stat.s.mode) {
      case 1: lcd.print(F("Trip 1"));            break;
      case 2: lcd.print(F("Trip 2"));            break;
      case 3: lcd.print(F("Trip 3"));            break;
      case 4: lcd.print(F("Trip 4"));            break;
      case 7: lcd.print(F("Altitude"));          break;
      case 8: lcd.print(F("Speed"));             break;
      case 9: lcd.print(F("Latitude"));          break;
      case 10: lcd.print(F("Longitude"));        break;
      case 12: lcd.print(F("GPS Satellites"));   break;
      case 13: lcd.print(F("GPS HDOP"));         break;
      case 14: lcd.print(F("Backlight"));        break;
      case 15: lcd.print(F("Runtime"));          break;
      case 16: lcd.print(F("Total Runtime"));           break;
      case 17: lcd.print(F("Battery"));          break;
      case 18: lcd.print(F("7 Segment"));        break;
      case 19: lcd.print(F("Last Lat")); lcd.setCursor(0, 1); lcd.print(opening_last_lat1, 7); lcd.write(223);     break;
      case 20: lcd.print(F("Last Lng")); lcd.setCursor(0, 1); lcd.print(opening_last_lng1, 7); lcd.write(223);     break;
      case 22: lcd.print(F("Dist s/pwron"));     break;
      case 23: lcd.print(F("High Score"));       break;
    }
  }

  if (millis() >= lastMillis + 1000) {
    lastMillis = millis();
    saveSettings();
    disp7((byte)round(gps.speed.mph()));

    switch (stat.s.mode) {
      case 0:
        lcd.clear(); lcd.print(gps.course.deg(), 1); lcd.write(223); lcd.print(' '); lcd.print(TinyGPSPlus::cardinal(gps.course.deg()));
        lcd.setCursor(0, 1);
        for (byte i = 4; i != 0; i--) {
          if ((float)stat.s.odometer / 1609340000 < pow(10, i)) {
            lcd.print('0');
          }
        }
        lcd.print((float)stat.s.odometer / 1609340000, 2);
        lcd.setCursor(12, 1);
        if (gps.speed.mph() < 10) {
          lcd.print('0');
        }
        lcd.print(gps.speed.mph());
        break;
      case 1: line2clear(); lcd.print((float)stat.s.trip1 / 1609340000, 3);        break;
      case 2: line2clear(); lcd.print((float)stat.s.trip2 / 1609340000, 3);        break;
      case 3: line2clear(); lcd.print((float)stat.s.trip3 / 1609340000, 3);        break;
      case 4: line2clear(); lcd.print((float)stat.s.trip4 / 1609340000, 3);        break;
      case 5: //CURRENT POSITION
        lcd.clear();
        lcd.print(gps.location.lat(), 7); lcd.write(223); lcd.setCursor(0, 1);
        lcd.print(gps.location.lng(), 7); lcd.write(223);
        break;
      case 6://UTC CLOCK
        lcd.clear();
        lcd.print(gps.date.year());  lcd.print('/'); lcd.print(gps.date.month()); lcd.print('/'); lcd.print(gps.date.day());
        lcd.setCursor(0, 1);
        if (gps.time.hour() < 10) {
          lcd.print('0');
        } lcd.print(gps.time.hour());  lcd.print(':');
        if (gps.time.minute() < 10) {
          lcd.print('0');
        } lcd.print(gps.time.minute());  lcd.print(':');
        if (gps.time.second() < 10) {
          lcd.print('0');
        } lcd.print(gps.time.second());
        break;
      case 7: line2clear(); lcd.print(gps.altitude.feet()); lcd.print(F(" ft"));        break;
      case 8: line2clear(); lcd.print(gps.speed.mph()); lcd.print(F(" mph"));           break;
      case 9: line2clear();  printCoord(stat.s.lat1);        break;
      case 10: line2clear(); printCoord(stat.s.lng1);        break;
      case 11: distang(stat.s.lat1, stat.s.lng1); break;
      case 12: line2clear(); lcd.print(gps.satellites.value()); break;
      case 13: line2clear(); lcd.print(gps.hdop.value());       break;
      case 14: line2clear(); lcd.print(stat.s.backlt);          break;
      case 15: line2clear(); lcd.print(millis() / 1000);       break;
      case 16: line2clear(); lcd.print(stat.s.hourmeter);               break;
      case 17: line2clear(); lcd.print(bvolt, 3); lcd.print(F(" V"));        break;
      case 18: line2clear(); lcd.print(stat.s.sevseg);          break;
      //case 19
      //case 20
      case 21: distang(opening_last_lat1, opening_last_lng1); break;
      case 22: line2clear(); lcd.print((float)tripPwrOn / 1609340000, 3);        break;
      case 23: line2clear(); lcd.print(highscore, 3);         break;
      
    }
  }

  if (currentKey == 1) { //MODE BUTTON PRESSED
    lcd.cursor();
    switch (stat.s.mode) {
      case 1: case 2: case 3: case 4: {
          byte k = 0;
          do {
            smartDelay(10); while (userButtons() != 0) {}; smartDelay(10);
            k = 0; while (k == 0) {
              k = userButtons();
            }
            if (k == 8) {
              switch (stat.s.mode) {
                case 1: stat.s.trip1 = 0; break;
                case 2: stat.s.trip2 = 0; break;
                case 3: stat.s.trip3 = 0; break;
                case 4: stat.s.trip4 = 0; break;
              }
              line2clear(); lcd.print(F("0.000"));
            }
          } while (k != 1);
        } break;
      case 9:  stat.s.lat1 = enterCoord(stat.s.lat1); break;
      case 10: stat.s.lng1 = enterCoord(stat.s.lng1); break;
      case 14: stat.s.backlt = enterByte(stat.s.backlt); break;
      case 18: stat.s.sevseg = enterByte(stat.s.sevseg); break;
    }

    lcd.noCursor();
    applySettings();
    saveSettings();
  }

  if (currentKey != 0) {
    while (userButtons() != 0) {}
  }
}
